# -*- coding: utf-8 -*-


class UnknownTransformation(ValueError):
    pass


class KeySignalNameError(ValueError):
    pass


class SignalNotFoundError(ValueError):
    pass


class AppNodeNotProvidedError(ValueError):
    pass
